
local MAJOR, MINOR = "LibJSON", 0.1
local LibJSON = LibStub:NewLibrary(MAJOR, MINOR)

-----------------------------------------------------------------------------
-- Module declaration
-----------------------------------------------------------------------------


do
   -----------------------------------------------------------------------------
   -- Imports and dependencies
   -----------------------------------------------------------------------------
   local math_floor, math_max, math_type = math.floor, math.max, math.type or function() end
   local string_char, string_sub, string_find, string_match, string_gsub, string_format
      = string.char, string.sub, string.find, string.match, string.gsub, string.format
   local table_insert, table_remove, table_concat = table.insert, table.remove, table.concat
   local type, tostring, pairs, assert, error = type, tostring, pairs, assert, error
   local loadstring = loadstring or load

   --- The LibJSON.null table allows one to specify a null value in an associative array (which is otherwise
   -- discarded if you set the value with 'nil' in Lua. Simply set t = { first=LibJSON.null }
   local null = {"This Lua table is used to designate JSON null value, compare your values with LibJSON.null to determine JSON nulls"}
   LibJSON.null = setmetatable(null, {
      __tostring = function() return 'null' end
   })

   --- The LibJSON.empty table allows one to specify an empty JSON object.
   -- To encode empty JSON array use usual empty Lua table.
   -- Example: t = { empty_object=LibJSON.empty, empty_array={} }
   local empty = {}
   LibJSON.empty = setmetatable(empty, {
      __tostring = function() return '{}' end,
      __newindex = function() error("LibJSON.empty is an read-only Lua table", 2) end
   })

   -----------------------------------------------------------------------------
   -- Private functions
   -----------------------------------------------------------------------------
   local decode
   local decode_scanArray
   local decode_scanConstant
   local decode_scanNumber
   local decode_scanObject
   local decode_scanString
   local decode_scanIdentifier
   local decode_scanWhitespace
   local encodeString
   local isArray
   local isEncodable
   local isConvertibleToString
   local isRegularNumber

   -----------------------------------------------------------------------------
   -- PUBLIC FUNCTIONS
   -----------------------------------------------------------------------------
   --- Encodes an arbitrary Lua object / variable.
   -- @param   obj     Lua value (table/string/boolean/number/nil/LibJSON.null/LibJSON.empty) to be JSON-encoded.
   -- @return  string  String containing the JSON encoding.
   function LibJSON.encode(obj)
      -- Handle nil and null values
      if obj == nil or obj == null then
         return 'null'
      end

      -- Handle empty JSON object
      if obj == empty then
         return '{}'
      end

      local obj_type = type(obj)

      -- Handle strings
      if obj_type == 'string' then
         return '"'..encodeString(obj)..'"'
      end

      -- Handle booleans
      if obj_type == 'boolean' then
         return tostring(obj)
      end

      -- Handle numbers
      if obj_type == 'number' then
         assert(isRegularNumber(obj), 'numeric values Inf and NaN are unsupported')
         return math_type(obj) == 'integer' and tostring(obj) or string_format('%.17g', obj)
      end

      -- Handle tables
      if obj_type == 'table' then
         local rval = {}
         -- Consider arrays separately
         local bArray, maxCount = isArray(obj)
         if bArray then
            for i = obj[0] ~= nil and 0 or 1, maxCount do
               table_insert(rval, LibJSON.encode(obj[i]))
            end
         else  -- An object, not an array
            for i, j in pairs(obj) do
               if isConvertibleToString(i) and isEncodable(j) then
                  table_insert(rval, '"'..encodeString(i)..'":'..LibJSON.encode(j))
               end
            end
         end
         if bArray then
            return '['..table_concat(rval, ',')..']'
         else
            return '{'..table_concat(rval, ',')..'}'
         end
      end

      error('Unable to JSON-encode Lua value of unsupported type "'..obj_type..'": '..tostring(obj))
   end

   local function create_state(s)
      -- Argument s may be "whole JSON string" or "JSON loader function"
      -- Returns "state" object which holds current state of reading long JSON:
      --    part = current part (substring of long JSON string)
      --    disp = number of bytes before current part inside long JSON
      --    more = function to load next substring (more == nil if all substrings are already read)
      local state = {disp = 0}
      if type(s) == "string" then
         -- s is whole JSON string
         state.part = s
      else
         -- s is loader function
         state.part = ""
         state.more = s
      end
      return state
   end

   --- Decodes a JSON string and returns the decoded value as a Lua data structure / value.
   -- @param   s           The string to scan (or "loader function" for getting next substring).
   -- @param   pos         (optional) The position inside s to start scan, default = 1.
   -- @return  Lua object  The object that was scanned, as a Lua table / string / number / boolean / LibJSON.null / LibJSON.empty.
   function LibJSON.decode(s, pos)
      return (decode(create_state(s), pos or 1))
   end

   --- Traverses a JSON string, sends everything to user-supplied callback function, returns nothing
   -- @param   s           The string to scan (or "loader function" for getting next substring).
   -- @param   callback    The user-supplied callback function which accepts arguments (path, json_type, value, pos, pos_last).
   -- @param   pos         (optional) The position inside s to start scan, default = 1.
   function LibJSON.traverse(s, callback, pos)
      decode(create_state(s), pos or 1, {path = {}, callback = callback})
   end

   local function read_ahead(state, startPos)
      -- Make sure there are at least 32 bytes read ahead
      local endPos = startPos + 31
      local part = state.part  -- current part (substring of "whole JSON" string)
      local disp = state.disp  -- number of bytes before current part inside "whole JSON" string
      local more = state.more  -- function to load next substring
      assert(startPos > disp)
      while more and disp + #part < endPos do
         --  (disp + 1) ... (disp + #part)  -  we already have this segment now
         --  startPos   ... endPos          -  we need to have this segment
         local next_substr = more()
         if not next_substr or next_substr == "" then
            more = nil
         else
            disp, part = disp + #part, string_sub(part, startPos - disp)
            disp, part = disp - #part, part..next_substr
         end
      end
      state.disp, state.part, state.more = disp, part, more
   end

   local function get_word(state, startPos, length)
      -- 1 <= length <= 32
      if state.more then read_ahead(state, startPos) end
      local idx = startPos - state.disp
      return string_sub(state.part, idx, idx + length - 1)
   end

   local function skip_until_word(state, startPos, word)
      -- #word < 30
      -- returns position after that word (nil if not found)
      repeat
         if state.more then read_ahead(state, startPos) end
         local part, disp = state.part, state.disp
         local b, e = string_find(part, word, startPos - disp, true)
         if b then
            return disp + e + 1
         end
         startPos = disp + #part + 2 - #word
      until not state.more
   end

   local function match_with_pattern(state, startPos, pattern, operation)
      -- pattern must be
      --    "^[some set of chars]+"
      -- returns
      --    matched_string, endPos   for operation "read"   (matched_string == "" if no match found)
      --    endPos                   for operation "skip"
      if operation == "read" then
         local t = {}
         repeat
            if state.more then read_ahead(state, startPos) end
            local part, disp = state.part, state.disp
            local str = string_match(part, pattern, startPos - disp)
            if str then
               table_insert(t, str)
               startPos = startPos + #str
            end
         until not str or startPos <= disp + #part
         return table_concat(t), startPos
      elseif operation == "skip" then
         repeat
            if state.more then read_ahead(state, startPos) end
            local part, disp = state.part, state.disp
            local b, e = string_find(part, pattern, startPos - disp)
            if b then
               startPos = startPos + e - b + 1
            end
         until not b or startPos <= disp + #part
         return startPos
      else
         error("Wrong operation name")
      end
   end

   --- Decodes a JSON string and returns the decoded value as a Lua data structure / value.
   -- @param   state             The state of JSON reader.
   -- @param   startPos          Starting position where the JSON string is located.
   -- @param   traverse          (optional) table with fields "path" and "callback" for traversing JSON.
   -- @param   decode_key        (optional) boolean flag for decoding key inside JSON object.
   -- @return  Lua_object,int    The object that was scanned, as a Lua table / string / number / boolean / LibJSON.null / LibJSON.empty,
   --                            and the position of the first character after the scanned JSON object.
   function decode(state, startPos, traverse, decode_key)
      local curChar, value, nextPos
      startPos, curChar = decode_scanWhitespace(state, startPos)
      if curChar == '{' and not decode_key then
         -- Object
         if traverse and traverse.callback(traverse.path, "object", nil, startPos, nil) then
            -- user wants to decode this JSON object (and get it as Lua value) while traversing
            local object, endPos = decode_scanObject(state, startPos)
            traverse.callback(traverse.path, "object", object, startPos, endPos - 1)
            return false, endPos
         end
         return decode_scanObject(state, startPos, traverse)
      elseif curChar == '[' and not decode_key then
         -- Array
         if traverse and traverse.callback(traverse.path, "array", nil, startPos, nil) then
            -- user wants to decode this JSON array (and get it as Lua value) while traversing
            local array, endPos = decode_scanArray(state, startPos)
            traverse.callback(traverse.path, "array", array, startPos, endPos - 1)
            return false, endPos
         end
         return decode_scanArray(state, startPos, traverse)
      elseif curChar == '"' then
         -- String
         value, nextPos = decode_scanString(state, startPos)
         if traverse then
            traverse.callback(traverse.path, "string", value, startPos, nextPos - 1)
         end
      elseif decode_key then
         -- Unquoted string as key name
         return decode_scanIdentifier(state, startPos)
      elseif string_find(curChar, "^[%d%-]") then
         -- Number
         value, nextPos = decode_scanNumber(state, startPos)
         if traverse then
            traverse.callback(traverse.path, "number", value, startPos, nextPos - 1)
         end
      else
         -- Otherwise, it must be a constant
         value, nextPos = decode_scanConstant(state, startPos)
         if traverse then
            traverse.callback(traverse.path, value == null and "null" or "boolean", value, startPos, nextPos - 1)
         end
      end
      return value, nextPos
   end

   -----------------------------------------------------------------------------
   -- Internal, PRIVATE functions.
   -- Following a Python-like convention, I have prefixed all these 'PRIVATE'
   -- functions with an underscore.
   -----------------------------------------------------------------------------

   --- Scans an array from JSON into a Lua object
   -- startPos begins at the start of the array.
   -- Returns the array and the next starting position
   -- @param   state       The state of JSON reader.
   -- @param   startPos    The starting position for the scan.
   -- @param   traverse    (optional) table with fields "path" and "callback" for traversing JSON.
   -- @return  table,int   The scanned array as a table, and the position of the next character to scan.
   function decode_scanArray(state, startPos, traverse)
      local array = not traverse and {}  -- The return value
      local elem_index, elem_ready, object = 1
      startPos = startPos + 1
      -- Infinite loop for array elements
      while true do
         repeat
            local curChar
            startPos, curChar = decode_scanWhitespace(state, startPos)
            if curChar == ']' then
               return array, startPos + 1
            elseif curChar == ',' then
               if not elem_ready then
                  -- missing value in JSON array
                  if traverse then
                     table_insert(traverse.path, elem_index)
                     traverse.callback(traverse.path, "null", null, startPos, startPos - 1)  -- empty substring: pos_last = pos - 1
                     table_remove(traverse.path)
                  else
                     array[elem_index] = null
                  end
               end
               elem_ready = false
               elem_index = elem_index + 1
               startPos = startPos + 1
            end
         until curChar ~= ','
         if elem_ready then
            error('Comma is missing in JSON array at position '..startPos)
         end
         if traverse then
            table_insert(traverse.path, elem_index)
         end
         object, startPos = decode(state, startPos, traverse)
         if traverse then
            table_remove(traverse.path)
         else
            array[elem_index] = object
         end
         elem_ready = true
      end
   end

   --- Scans for given constants: true, false or null
   -- Returns the appropriate Lua type, and the position of the next character to read.
   -- @param  state        The state of JSON reader.
   -- @param  startPos     The position in the string at which to start scanning.
   -- @return object, int  The object (true, false or LibJSON.null) and the position at which the next character should be scanned.
   function decode_scanConstant(state, startPos)
      local w5 = get_word(state, startPos, 5)
      local w4 = string_sub(w5, 1, 4)
      if w5 == "false" then
         return false, startPos + 5
      elseif w4 == "true" then
         return true, startPos + 4
      elseif w4 == "null" then
         return null, startPos + 4
      end
      error('Failed to parse JSON at position '..startPos)
   end

   --- Scans a number from the JSON encoded string.
   -- (in fact, also is able to scan numeric +- eqns, which is not in the JSON spec.)
   -- Returns the number, and the position of the next character after the number.
   -- @param   state        The state of JSON reader.
   -- @param   startPos     The position at which to start scanning.
   -- @return  number,int   The extracted number and the position of the next character to scan.
   function decode_scanNumber(state, startPos)
      local stringValue, endPos = match_with_pattern(state, startPos, '^[%+%-%d%.eE]+', "read")
      local stringEval = loadstring('return '..stringValue)
      if not stringEval then
         error('Failed to scan number '..stringValue..' in JSON string at position '..startPos)
      end
      return stringEval(), endPos
   end

   --- Scans a JSON object into a Lua object.
   -- startPos begins at the start of the object.
   -- Returns the object and the next starting position.
   -- @param   state       The state of JSON reader.
   -- @param   startPos    The starting position of the scan.
   -- @param   traverse    (optional) table with fields "path" and "callback" for traversing JSON
   -- @return  table,int   The scanned object as a table and the position of the next character to scan.
   function decode_scanObject(state, startPos, traverse)
      local object, elem_ready = not traverse and empty
      startPos = startPos + 1
      while true do
         repeat
            local curChar
            startPos, curChar = decode_scanWhitespace(state, startPos)
            if curChar == '}' then
               return object, startPos + 1
            elseif curChar == ',' then
               startPos = startPos + 1
               elem_ready = false
            end
         until curChar ~= ','
         if elem_ready then
            error('Comma is missing in JSON object at '..startPos)
         end
         -- Scan the key as string or unquoted identifier such as in {"a":1,b:2}
         local key, value
         key, startPos = decode(state, startPos, nil, true)
         local colon
         startPos, colon = decode_scanWhitespace(state, startPos)
         if colon ~= ':' then
            error('JSON object key-value assignment mal-formed at '..startPos)
         end
         startPos = decode_scanWhitespace(state, startPos + 1)
         if traverse then
            table_insert(traverse.path, key)
         end
         value, startPos = decode(state, startPos, traverse)
         if traverse then
            table_remove(traverse.path)
         else
            if object == empty then
               object = {}
            end
            object[key] = value
         end
         elem_ready = true
      end  -- infinite loop while key-value pairs are found
   end

   --- Scans JSON string for an identifier (unquoted key name inside object)
   -- Returns the string extracted as a Lua string, and the position after the closing quote.
   -- @param  state        The state of JSON reader.
   -- @param  startPos     The starting position of the scan.
   -- @return string,int   The extracted string as a Lua string, and the next character to parse.
   function decode_scanIdentifier(state, startPos)
      local identifier, idx = match_with_pattern(state, startPos, '^[%w_%-%$]+', "read")
      if identifier == "" then
         error('JSON String decoding failed: missing key name at position '..startPos)
      end
      return identifier, idx
   end

   -- START SoniEx2
   -- Initialize some things used by decode_scanString
   -- You know, for efficiency
   local escapeSequences = { t = "\t", f = "\f", r = "\r", n = "\n", b = "\b" }
   -- END SoniEx2

   --- Scans a JSON string from the opening quote to the end of the string.
   -- Returns the string extracted as a Lua string, and the position after the closing quote.
   -- @param  state        The state of JSON reader.
   -- @param  startPos     The starting position of the scan.
   -- @return string,int   The extracted string as a Lua string, and the next character to parse.
   function decode_scanString(state, startPos)
      local t, idx, surrogate_pair_started, regular_part = {}, startPos + 1
      while true do
         regular_part, idx = match_with_pattern(state, idx, '^[^"\\]+', "read")
         table_insert(t, regular_part)
         local w6 = get_word(state, idx, 6)
         local c = string_sub(w6, 1, 1)
         if c == '"' then
            return table_concat(t), idx + 1
         elseif c == '\\' then
            local esc = string_sub(w6, 2, 2)
            if esc == "u" then
               local n = tonumber(string_sub(w6, 3), 16)
               if not n then
                  error("String decoding failed: bad Unicode escape "..w6.." at position "..idx)
               end
               -- Handling of UTF-16 surrogate pairs
               if n >= 0xD800 and n < 0xDC00 then
                  surrogate_pair_started, n = n
               elseif n >= 0xDC00 and n < 0xE000 then
                  n, surrogate_pair_started = surrogate_pair_started and (surrogate_pair_started - 0xD800) * 0x400 + (n - 0xDC00) + 0x10000
               end
               if n then
                  -- Convert unicode codepoint n (0..0x10FFFF) to UTF-8 string
                  local x
                  if n < 0x80 then
                     x = string_char(n % 0x80)
                  elseif n < 0x800 then
                     -- [110x xxxx] [10xx xxxx]
                     x = string_char(0xC0 + (math_floor(n/64) % 0x20), 0x80 + (n % 0x40))
                  elseif n < 0x10000 then
                     -- [1110 xxxx] [10xx xxxx] [10xx xxxx]
                     x = string_char(0xE0 + (math_floor(n/64/64) % 0x10), 0x80 + (math_floor(n/64) % 0x40), 0x80 + (n % 0x40))
                  else
                     -- [1111 0xxx] [10xx xxxx] [10xx xxxx] [10xx xxxx]
                     x = string_char(0xF0 + (math_floor(n/64/64/64) % 8), 0x80 + (math_floor(n/64/64) % 0x40), 0x80 + (math_floor(n/64) % 0x40), 0x80 + (n % 0x40))
                  end
                  table_insert(t, x)
               end
               idx = idx + 6
            else
               table_insert(t, escapeSequences[esc] or esc)
               idx = idx + 2
            end
         else
            error('String decoding failed: missing closing " for string at position '..startPos)
         end
      end
   end

   --- Scans a JSON string skipping all whitespace from the current start position.
   -- Returns the position of the first non-whitespace character.
   -- @param   state      The state of JSON reader.
   -- @param   startPos   The starting position where we should begin removing whitespace.
   -- @return  int,char   The first position where non-whitespace was encountered, non-whitespace char.
   function decode_scanWhitespace(state, startPos)
      while true do
         startPos = match_with_pattern(state, startPos, '^[ \n\r\t]+', "skip")
         local w2 = get_word(state, startPos, 2)
         if w2 == '/*' then
            local endPos = skip_until_word(state, startPos + 2, '*/')
            if not endPos then
               error("Unterminated comment in JSON string at "..startPos)
            end
            startPos = endPos
         else
            local next_char = string_sub(w2, 1, 1)
            if next_char == '' then
               error('Unexpected end of JSON')
            end
            return startPos, next_char
         end
      end
   end

   --- Encodes a string to be JSON-compatible.
   -- This just involves backslash-escaping of quotes, slashes and control codes
   -- @param   s        The string to return as a JSON encoded (i.e. backquoted string)
   -- @return  string   The string appropriately escaped.
   local escapeList = {
         ['"']  = '\\"',
         ['\\'] = '\\\\',
         ['/']  = '\\/',
         ['\b'] = '\\b',
         ['\f'] = '\\f',
         ['\n'] = '\\n',
         ['\r'] = '\\r',
         ['\t'] = '\\t',
         ['\127'] = '\\u007F'
   }
   function encodeString(s)
      if type(s) == 'number' then
         s = math_type(s) == 'integer' and tostring(s) or string_format('%.f', s)
      end
      return string_gsub(s, ".", function(c) return escapeList[c] or c:byte() < 32 and string_format('\\u%04X', c:byte()) end)
   end

   -- Determines whether the given Lua type is an array or a table / dictionary.
   -- We consider any table an array if it has indexes 1..n for its n items, and no other data in the table.
   -- I think this method is currently a little 'flaky', but can't think of a good way around it yet...
   -- @param   t                 The table to evaluate as an array
   -- @return  boolean,number    True if the table can be represented as an array, false otherwise.
   --                            If true, the second returned value is the maximum number of indexed elements in the array.
   function isArray(t)
      -- Next we count all the elements, ensuring that any non-indexed elements are not-encodable
      -- (with the possible exception of 'n')
      local maxIndex = 0
      for k, v in pairs(t) do
         if type(k) == 'number' and math_floor(k) == k and 0 <= k and k <= 1e6 then  -- k,v is an indexed pair
            if not isEncodable(v) then  -- All array elements must be encodable
               return false
            end
            maxIndex = math_max(maxIndex, k)
         elseif not (k == 'n' and v == #t) then  -- if it is n, then n does not hold the number of elements
            if isConvertibleToString(k) and isEncodable(v) then
               return false
            end
         end -- End of k,v not an indexed pair
      end  -- End of loop across all pairs
      return true, maxIndex
   end

   --- Determines whether the given Lua object / table / value can be JSON encoded.
   -- The only types that are JSON encodable are: string, boolean, number, nil, table and special tables LibJSON.null and LibJSON.empty.
   -- @param   o        The object to examine.
   -- @return  boolean  True if the object should be JSON encoded, false if it should be ignored.
   function isEncodable(o)
      local t = type(o)
      return t == 'string' or t == 'boolean' or t == 'number' and isRegularNumber(o) or t == 'nil' or t == 'table'
   end

   --- Determines whether the given Lua object / table / variable can be a JSON key.
   -- Integer Lua numbers are allowed to be considered as valid string keys in JSON.
   -- @param   o        The object to examine.
   -- @return  boolean  True if the object can be converted to a string, false if it should be ignored.
   function isConvertibleToString(o)
      local t = type(o)
      return t == 'string' or t == 'number' and isRegularNumber(o) and (math_type(o) == 'integer' or math_floor(o) == o)
   end

   local is_Inf_or_NaN = {[tostring(1/0)]=true, [tostring(-1/0)]=true, [tostring(0/0)]=true, [tostring(-(0/0))]=true}
   --- Determines whether the given Lua number is a regular number or Inf/Nan.
   -- @param   v        The number to examine.
   -- @return  boolean  True if the number is a regular number which may be encoded in JSON.
   function isRegularNumber(v)
      return not is_Inf_or_NaN[tostring(v)]
   end

end

return LibJSON